/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "sched.h"
#include "err_and_warn.h"
#include <process.h>

task*
task_pool::
store( const task_data& td )
{
	task* t = allocate(); if (!t) return NULL;
	t->td = td;
	mon.inc();
	return t;
}


task_pool scheduler::tp(1000);
task* scheduler::currentTask = NULL;
HANDLE scheduler::threadHandle;
UINT scheduler::threadID;
HANDLE scheduler::eventHandle;
CRITICAL_SECTION scheduler::crit;
Bool scheduler::running;

scheduler::
scheduler()
{
	InitializeCriticalSection(&crit);
	eventHandle = CreateEvent(NULL, FALSE, FALSE, "RMcB-sched");
	running = True;
	threadHandle = (HANDLE) _beginthreadex(NULL, 0, threadMain, NULL, 0, &threadID);
	SetThreadPriority(threadHandle, THREAD_PRIORITY_TIME_CRITICAL);
}

scheduler::
~scheduler()
{
	running = False;
	SetEvent(eventHandle); // wake up scheduler thread; let it see we're not running
	DWORD result = WaitForSingleObject(threadHandle, 1000); // wait for thread to die
	if (result == WAIT_TIMEOUT)
		ew.w("scheduler thread failed to exit");
	if (!CloseHandle(eventHandle))
		ew.w("failed to close eventHandle in scheduler cleanup");
	if (!CloseHandle(threadHandle))
		ew.w("failed to close threadHandle in scheduler cleanup");
	DeleteCriticalSection(&crit);
}

unsigned WINAPI
scheduler::
threadMain(LPVOID lpvThreadParm)
{
	synchronize();
	while (running)
	{
		task* doMeNext;
		task** doMeNext_prev;
		reorderTasks(&doMeNext, &doMeNext_prev);
		waitAndHandle(doMeNext, doMeNext_prev); // releases synchronization while waiting
	}
	unsynchronize();
	return 0;
}

void
scheduler::
reorderTasks(task** doMeNextp, task*** doMeNext_prevp)
{
	if (currentTask != NULL) // find out what should be scheduled next
	{
		task* min = (task*) currentTask;
		task* i;
		task* min_prev = NULL;
		task* i_prev = NULL;
#ifndef NDEBUG
		if (currentTask == NULL)
			ew.w("sanity check failed in scheduler -- CurrentTask is NULL");
#endif
		for (i = (task*) currentTask; i != NULL; i = i->chain)
		{
#ifndef NDEBUG
			if (i == NULL)
				ew.w("sanity check failed in scheduler -- i is NULL");
#endif
			// explaination behind why this is "<=" instead of "<" below:
			// more recently scheduled tasks will appear first, since we add to the head of the queue.
			// but among tasks which are scheduled for the exact same time, we want to execute first
			// the one which was scheduled first
			if (i->td.nwu <= min->td.nwu)
			{
				min = i;
				min_prev = i_prev;
			}
			i_prev = i;
		}
		// at this point, min points to the task which should be scheduled next
		// min_prev points to the task right before min, or NULL if min == currentTask
		// currentTask points to the first task in the queue
		// i_prev points to the last task in the queue
#ifndef NDEBUG
		if (min == NULL)
			ew.w("sanity check failed in scheduler -- min is NULL");
#endif
//		if (min_prev)
//		{
//			min_prev->chain = NULL;
//			i_prev->chain = (task*) currentTask;
//			currentTask = min;
//		}
		*doMeNextp = min;
		if (min_prev == NULL)
			*doMeNext_prevp = &currentTask;
		else
			*doMeNext_prevp = &(min_prev->chain);
	}
}

void
scheduler::
waitAndHandle(task* doMeNext, task** doMeNext_prev)
{
	DWORD dwTimeout;
	if (currentTask != NULL)
	{
		long delay = currentTask->td.nwu - oap->time(); // calculate time for wakeup of next task
		dwTimeout = ((delay > 0) ? (delay/2) : 0);
	}
	else
		dwTimeout = INFINITE; // no task scheduled; delay forever (until we get MIDI input)
	unsynchronize();
	DWORD result = WaitForSingleObject(eventHandle, dwTimeout);
	synchronize();
	if (result == WAIT_TIMEOUT)
		removeAndExecute(doMeNext, doMeNext_prev); // execute scheduled task
	else
		oap->processQueue(); // process MIDI input
}

void
scheduler::
processQueue()
{
	SetEvent(eventHandle);
}

void
scheduler::
removeAndExecute(task* doMeNext, task** doMeNext_prev)
{
	ScheduleTime oaptime = oap->time();
	if (currentTask == NULL || oaptime < currentTask->td.nwu) // this might happen if main thread 
		return;                                // removed a scheduled event while we were sleeping

//	task* t = currentTask;
//	task_data& td = currentTask->td;
//	currentTask = currentTask->chain; // remove task
	task* t = doMeNext;
	task_data& td = doMeNext->td;
	*doMeNext_prev = doMeNext->chain;

	if (statistics_on)
		tp.record_lateness( oaptime - td.nwu );
	
	// execute task
	Bool retval = False;
	switch (td.ty)
	{
		case task_data::fa_t:
		{
			long r_time = td.fa.func( td.fa.arg );
			if (r_time != NEVER)
			{
				td.nwu += r_time;
				retval = True;
			}
			break;
		}
		case task_data::m_t: oap->write( &(td.m) ); break;
		case task_data::on_t:
			oap->write( &(td.m) );
			td.ty = task_data::m_t;
			td.m.d2 = 0;
			td.nwu += td.m.time;
			retval = True;
			break;
	}

	if (retval == False)
	{
		tp.mon.dec();
		tp.release( t );
	}
	else // implement ZOR
	{
		do_insert(t);
	}
}

err_ind
scheduler::
play_note( const MIDI_msg& on, long dur )
{
	oap->write( &on );
	MIDI_msg off = on;
	off.d2 = 0;
	task_data td( off, oap->time() + dur );
	return insert(td);
}

err_ind
scheduler::
do_insert(task* t)
{
	t->chain = currentTask;
	currentTask = t;
	if (GetCurrentThreadId() != threadID) // this hack is so that if we schedule from another thread,
		SetEvent(eventHandle);             // we will recompute which event is scheduled next
	return no_err;
}

err_ind
scheduler::
cancel(task* t)
{
	synchronize();
	if (t == currentTask)
	{
		currentTask = t->chain;
		tp.mon.dec();
		tp.release(t);
		unsynchronize();
		return no_err;
	}
	task* i;
	task* i_prev = NULL;
	// won't hit first time through loop because we know t != currentTask
	for (i = (task*) currentTask; i != NULL; i = i->chain)
	{
		if (i == t)
		{
			i_prev->chain = i->chain; // remove link
			tp.mon.dec();
			tp.release( t );
			unsynchronize();
			return no_err;
		}
		i_prev = i;
	}
	ew.w("failed attempt to cancel task -- Event not found!");
	unsynchronize();
	return err;
}

//--------------------------------------------------------------------------------------------------

// Authors: Ben Denckla, Patrick Pelletier
