/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "text_win.h"
#include "Rogus_McBogus.h"
#include <string.h>
#include <stdio.h> // for vsprintf
#include "err_and_warn.h"
#include <process.h> // for _beginthreadex

text_win tw;

void
text_win::
my_outfunc(const char* msg, int msg_type)
{
	HANDLE hStdout;
	CONSOLE_SCREEN_BUFFER_INFO info;
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hStdout, &info); // save old color
	switch (msg_type)
	{
	case Rogus_McBogus::msg:
		SetConsoleTextAttribute(hStdout, FOREGROUND_GREEN | FOREGROUND_INTENSITY);
		break;
	case Rogus_McBogus::warn:
		SetConsoleTextAttribute(hStdout, FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY);
		break;
	case Rogus_McBogus::err:
		SetConsoleTextAttribute(hStdout, FOREGROUND_RED | FOREGROUND_INTENSITY);
		break;
	}
	DWORD numchars = strlen(msg);
	WriteConsole(hStdout, (const void*) msg, numchars, &numchars, NULL);
	SetConsoleTextAttribute(hStdout, info.wAttributes); // restore old color
}

text_win::
~text_win()
{
	if (inited)
	{
		running = False;
		SetEvent(eventHandle); // wake up output thread; let it see we're not running
		DWORD result = WaitForSingleObject(threadHandle, 1000); // wait for thread to die
		if (result == WAIT_TIMEOUT)
			outfunc("Warning: text_win thread failed to exit\n",Rogus_McBogus::warn);
		if (!CloseHandle(eventHandle))
			outfunc("Warning: failed to close eventHandle in text_win cleanup\n",Rogus_McBogus::warn);
		if (!CloseHandle(threadHandle))
			outfunc("Warning: failed to close threadHandle in text_win cleanup\n",Rogus_McBogus::warn);
		DeleteCriticalSection(&crit);
		inited = False;
	}
}

unsigned WINAPI
text_win::
threadMain(LPVOID lpvThreadParm)
{
	while (tw.running)
	{
		WaitForSingleObject(tw.eventHandle, INFINITE);
		while (tw.len > 0)
		{
			unit* uu = &tw.units[tw.pos];
			if (tw.len > 1)
			{
				tw.outfunc(uu->buf, uu->msg_type);
				EnterCriticalSection(&tw.crit);
				tw.pos = (tw.pos + 1) % numunits;
				tw.len--;
				LeaveCriticalSection(&tw.crit);
			}
			else
			{
				EnterCriticalSection(&tw.crit);
				unit u = *uu;
				tw.pos = (tw.pos + 1) % numunits;
				tw.len--;
				LeaveCriticalSection(&tw.crit);
				tw.outfunc(u.buf, u.msg_type);
			}
		}
	}
	return 0;
}

void
text_win::
init(void (*func)(const char*, int))								
{
	if (!inited)
	{
		pos = 0;
		len = 0;
		InitializeCriticalSection(&crit);
		UINT threadID;
		eventHandle = CreateEvent(NULL, FALSE, FALSE, "RMcB-tw");
		running = True;
		threadHandle = (HANDLE) _beginthreadex(NULL, 0, threadMain, NULL, 0, &threadID);
		SetThreadPriority(threadHandle, THREAD_PRIORITY_LOWEST);
	}
	if (func != NULL)
	{
		outfunc = func;
		inited = True;
		return;
	}
	outfunc = my_outfunc;
	AllocConsole(); // create a console window if we don't already have one
	inited = True;
	// check if we are running under Windows 95
	// if so, don't add scrollbar, because Windows 95 console
	// windows have a serious bug when using scrollbars
	OSVERSIONINFO ovi;
	ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	if (!GetVersionEx(&ovi))
	{
		ew.w("GetVersionEx failed");
		return;
	}
	if (ovi.dwPlatformId != VER_PLATFORM_WIN32_NT)
	{
		tw.pr("Note: scrollbar not available under Windows 95\n");
		return;
	}
	// make sure console window has at least 200 lines to scroll back
	HANDLE hStdout;
	CONSOLE_SCREEN_BUFFER_INFO info;
	hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
	GetConsoleScreenBufferInfo(hStdout, &info);
	Bool dochange = False;
	if (info.dwSize.X < 80)
	{
		info.dwSize.X = 80;
		dochange = True;
	}
	if (info.dwSize.Y < 200)
	{
		info.dwSize.Y = 200;
		dochange = True;
	}
	if (dochange)
		SetConsoleScreenBufferSize(hStdout, info.dwSize);
}

void
text_win::
pr( const char *format, ... )
{
	va_list args;
	va_start( args, format );
	
	vpr( format, args, Rogus_McBogus::msg);
	
	va_end( args );
}	

void
text_win::
spr( const char *format, ... )
{
	va_list args;
	va_start( args, format );
	
	vpr( format, args, Rogus_McBogus::msg );
	
	va_end( args );
}	

void
text_win::
vpr( const char *format, va_list args, int msg_type)
{
	if (!inited)
		init(NULL);
	int l = vsprintf( buf, format, args );
	char* where = buf;
	Bool prichanged = False;
	int backoff = 1;
	while (l > 0)
	{
		EnterCriticalSection(&crit);
		unit* u1 = &units[(pos+len)%numunits];
		unit* uu = u1;
		if (len > 0) // try to stuff into previous one if possible
		{
			unit* u2 = &units[(pos+len-1)%numunits];
			if (u2->msg_type == msg_type && (u2->len + l) < bufsize)
				uu = u2;
		}
		if (uu == u1)
		{
			if (len == numunits) // running behind?
			{
				LeaveCriticalSection(&crit);
				SetThreadPriority(threadHandle, THREAD_PRIORITY_TIME_CRITICAL); // bump up priority
				prichanged = True;
				SetEvent(eventHandle); // for good measure
				Sleep(backoff);
				if (backoff < 1024)
					backoff *= 2;
				continue;
			}
			uu->len = 0;
			len++;
		}
		int start = uu->len;
		for (int i = 0; where[i] != 0 && i < bufsize - 1; i++)
		{
			if (where[i] == '\r')
				uu->buf[i+start] = '\n';
			else
				uu->buf[i+start] = where[i];
		}
		uu->buf[i+start] = 0;
		uu->len += i;
		uu->msg_type = msg_type;
		where += i;
		l -= i;
		LeaveCriticalSection(&crit);
	}

	if (prichanged)
		SetThreadPriority(threadHandle, THREAD_PRIORITY_LOWEST);

	SetEvent(eventHandle);
}

// Authors: Ben Denckla, Pete Rice, Patrick Pelletier
