/*
** Copyright 1995-1996 by the Massachusetts Institute of Technology.  All
** rights reserved. For more information, see the file COPYING which came
** with this source distribution.
*/

#include "NotePairScorePlayer.h"
#include "data_structs.h"

void NotePairScorePlayer::defaultOutFunc(const NotePairScoreElement& npse, NotePairScoreScale scale, ScheduleTime now)
{
	npse.play(scale, now);
}

NotePairScorePlayer::NotePairScorePlayer(NotePairScorePtr sp, void (*func)(const NotePairScoreElement&, NotePairScoreScale, ScheduleTime))
{
	score = sp;
	outfunc = func;
	playing = 0;
	muted = 0;
	looped = 0;
	scale = (NotePairScoreScale) 1;
	leftover = -1;
	index = 0;
}

NotePairScorePlayer::~NotePairScorePlayer()
{
	stop();
}

void NotePairScorePlayer::startMultiple(NotePairScorePlayerPtrArray* a)
{
	schedp->synchronize();
	ScheduleTime t = oap->time();
	int i;
	for (i = 0; i < a->length; i++)
	{
		NotePairScorePlayer* sp = (*a)[i];
		if (sp->playing || (sp->score->length == 0))
			continue;
		sp->do_start(t);
		sp->playing = 1;
	}
	schedp->unsynchronize();
}

void NotePairScorePlayer::start()
{
	schedp->synchronize();
	if (playing || (score->length == 0))
	{
		schedp->unsynchronize();
		return;
	}
	ScheduleTime t = oap->time();
	do_start(t);
	playing = 1;
	schedp->unsynchronize();
}

void NotePairScorePlayer::do_start(ScheduleTime tt)
{
	ScheduleTime t = tt;
	lastWakeUp = t;
	if (leftover != -1 && index > 0)
		t += (ScheduleTime) (scale*((*score)[index].msg.time-((*score)[index-1].msg.time+leftover)));
	nextWakeUp = t;
	if (nextWakeUp < lastWakeUp)
	{
//		tw.pr("We scheduled %ld demimilliseconds in the past\r",lastWakeUp-nextWakeUp);
		ew.w("NotePairScorePlayer is running %ld demimilliseconds late",lastWakeUp-nextWakeUp);
//		nextWakeUp = lastWakeUp;
//		leftover = -1;
	}
	task_data td(callback, (void*) this, nextWakeUp);
	if (schedp->insert(td, (task**)(&ott)) != no_err)
	{
		ew.w("schedp->insert failed");
		leftover = -1;
		playing = 0;
		return;
	}
}

void NotePairScorePlayer::stopMultiple(NotePairScorePlayerPtrArray* a)
{
	schedp->synchronize();
	ScheduleTime t = oap->time();
	int i;
	for (i = 0; i < a->length; i++)
	{
		NotePairScorePlayer* sp = (*a)[i];
		if (!(sp->playing))
			continue;
		if (sp->do_stop(t) == no_err)
		{
			sp->playing = 0;
		}
	}
	schedp->unsynchronize();
}

void NotePairScorePlayer::stop()
{
	schedp->synchronize();
	if (!playing)
	{
		schedp->unsynchronize();
		return;
	}
	ScheduleTime t = oap->time();
	if (do_stop(t) == no_err)
	{
		playing = 0;
	}
	schedp->unsynchronize();
}

err_ind NotePairScorePlayer::do_stop(ScheduleTime t)
{
	if (schedp->cancel((task*)ott) != no_err)
	{
		ew.w("schedp->cancel failed\r");
		return err;
	}
	if (leftover == -1)
		leftover = 0;
	leftover += (NotePairScoreTime) ((t-lastWakeUp)/scale);
	return no_err;
}

void NotePairScorePlayer::setIndex(int i)
{
	schedp->synchronize();
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
		{
			schedp->unsynchronize();
			return;
		}
	index = i;
	leftover = -1;
	if (playing)
		do_start(t);
	schedp->unsynchronize();
}

void NotePairScorePlayer::setTimeUnscaled(NotePairScoreTime x)
{
	schedp->synchronize();
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
		{
			schedp->unsynchronize();
			return;
		}
	index = 0;
	leftover = -1;
	if (score->length > 0)
		for (int i = 0; i < score->length; i++)
			if ((*score)[i].msg.time >= x)
			{
				if (i == 0 || (*score)[i].msg.time == x)
					leftover = -1;
				else
					leftover = (x - (*score)[i-1].msg.time);
				index = i;
				break;
			}
	if (playing)
		do_start(t);
	schedp->unsynchronize();
}

NotePairScoreTime NotePairScorePlayer::getTimeUnscaled() const
{
	schedp->synchronize();
	ScheduleTime t = oap->time();
	NotePairScoreTime lo = leftover;
	if (lo == -1)
		lo = 0;
	if (playing)
		lo += (NotePairScoreTime) ((t-lastWakeUp)/scale + 0.5);
	NotePairScoreTime base = ((index > 0) ? (*score)[index-1].msg.time : 0);
	schedp->unsynchronize();
	return base+lo;
}

void NotePairScorePlayer::setScale(NotePairScoreScale s)
{
	schedp->synchronize();
	ScheduleTime t = oap->time();
	if (playing)
		if (do_stop(t) == err)
		{
			schedp->unsynchronize();
			return;
		}
	scale = s;
	if (playing)
		do_start(t);
	schedp->unsynchronize();
}

void NotePairScorePlayer::setMute(Bool mu)
{
	schedp->synchronize();
	mu = ((mu == 1) ? 1 : 0);
	if (mu == muted)
	{
		schedp->unsynchronize();
		return;
	}
	ScheduleTime t = oap->time();
	if (playing && do_stop(t) == err)
	{
		schedp->unsynchronize();
		return;
	}
	muted = mu;
	if (playing)
		do_start(t);
	schedp->unsynchronize();
}

ScheduleTime NotePairScorePlayer::callback(void* dwUser)
{
	NotePairScorePlayer* p = (NotePairScorePlayer*) dwUser;
	NotePairScoreElement* npse = &((*(p->score))[p->index]);
	if (!(p->muted))
		p->outfunc(*npse, p->scale, p->nextWakeUp);
	p->index++;
	ScoreTime mdottime = npse->msg.time;
	if (p->index >= p->score->length)
	{
		p->index = 0;
		p->leftover = -1;
		if (!(p->looped))
		{
			p->playing = 0;
			return NEVER;
		}
		else
			mdottime = (*(p->score))[0].msg.time;
	}
	p->lastWakeUp = p->nextWakeUp;
//	tw.spr("m.time = %ld m.s = %d m.d1 = %d m.d2 = %d\r", m.time, m.s, m.d1, m.d2);
//	tw.spr("((*(p->score))[p->index]).time = %ld, m.time = %ld p->scale = %f\r",((*(p->score))[p->index]).time, m.time, p->scale);
	p->nextWakeUp += (ScheduleTime) (((((*(p->score))[p->index]).msg.time - mdottime) * p->scale) + 0.5);
//	tw.spr("We added %ld to nextWakeUp\r",p->nextWakeUp - p->lastWakeUp);
	p->leftover = -1;
	return (p->nextWakeUp - p->lastWakeUp);
}

// Author: Patrick Pelletier
